﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using AForge;
using AForge.Imaging;
using AForge.Imaging.Filters;
using AForge.Math.Geometry;
using System.Drawing.Imaging;

namespace KartenDetektor
{
    public partial class Form1 : Form
    {
        public Form1()
        {
            InitializeComponent();
        }

        //Liste für die Eingabe-Screenshots
        List<string> files = new List<string>();

        private void Form1_Load(object sender, EventArgs e)
        {
            //Dateinamen der Screenshots in die Liste laden
            files = System.IO.Directory.GetFiles(Application.StartupPath + "\\samples\\n").ToList();
            
            //Ersten Screenshot in Bitmap-Objekt laden 
            Bitmap inputBmp = (Bitmap)Bitmap.FromFile(files[0]);
            //Analyse durchführen
            analyzeCards(inputBmp);
            //Screenshot aus Liste entfernen
            files.RemoveAt(0);
        }


        public void analyzeCards(Bitmap inputBmp)
        {
            //Liste für Spielkarten-Objekte anlegen
            List<Card> cards = new List<Card>();

            //Screenshot auf Standardformat ziehen, damit unabhängig von der Auflösung
            //die nachfolgenden Koordinaten stimmen
            ResizeBilinear filterRes = new ResizeBilinear(1364, 726);
            inputBmp = filterRes.Apply(inputBmp);

            //Arbeitskopie des Screenshots anlegen
            Bitmap processedBmp = inputBmp.Clone() as Bitmap;

            //Farbfilter für die Arbeitskopie anlegen und anwenden, um 
            //bessere Ergebnisse bei der S/W-Konvertierung zu erhalten
            ColorFiltering filter = new ColorFiltering();
            filter.Red = new AForge.IntRange(250, 255);
            filter.Green = new AForge.IntRange(250, 255);
            filter.Blue = new AForge.IntRange(250, 255);
            filter.ApplyInPlace(processedBmp);

            //Arbeitskopie über Graustaufen nach S/W konvertieren
            FiltersSequence seq = new FiltersSequence();
            seq.Add(Grayscale.CommonAlgorithms.BT709);
            seq.Add(new IterativeThreshold()); 
            processedBmp = seq.Apply(processedBmp);

            //Spielkarten per Blobcounter finden. Durch MinHeight und MaxHeight wird die 
            //Größe der gefunden Blobs so eingeschränkt, dass nur Karten gefunden werden
            BlobCounter extractor = new BlobCounter();
            extractor.FilterBlobs = true;
            extractor.MinWidth = extractor.MinHeight = 40;
            extractor.MaxWidth = extractor.MaxHeight = 70;
            extractor.ProcessImage(processedBmp);
            var blobs = extractor.GetObjectsInformation().ToList();

            //Die ersten beiden Blobs entfernen (da dies die zwei verdeckten Karten
            //auf dem Screenshot sind.
            blobs.RemoveRange(0, 2);

            //Graphics-Object vom Screenshot (nicht der S/W-Arbeitskopie) anlegen,
            //um die gefunden Blobs (=Spielkarten) einzeichnen zu können
            Graphics gfx = Graphics.FromImage(inputBmp);

            //Durch alle Blobs (=gefunden Kartenpositionen) iterieren
            foreach (Blob blob in blobs)
            {
                //Eckpunkte des Spielkarten-Blobs in Liste schreiben
                List<IntPoint> edgePoints = extractor.GetBlobsEdgePoints(blob);
                //Spielkarten-Ecken finden
                List<IntPoint> corners = PointsCloud.FindQuadrilateralCorners(edgePoints);
                                
                //Spielkarte in Rechteck tranformieren/entzerren
                QuadrilateralTransformation quadTransformer = new QuadrilateralTransformation();
                quadTransformer.SourceQuadrilateral = corners;
                quadTransformer.AutomaticSizeCalculaton = true;
                Bitmap tempImg = quadTransformer.Apply(inputBmp); 

                //Wenn die Spielkarte auf der Seite liegt, Karte drehen/aufrichten
                if (tempImg.Width > tempImg.Height) 
                    tempImg.RotateFlip(RotateFlipType.Rotate270FlipNone); 

                //Spielkarte auf Standardmaße strecken, damit die nachfolgenden
                //Koordinaten auf alle Spielkarten-Grafiken anwendbar sind
                ResizeBilinear resizer = new ResizeBilinear(200, 300);
                tempImg = resizer.Apply(tempImg); 

                //Neue Spielkarten-Objekt anlegen und die Spielkarten-Grafik hinzufügen
                var card = new Card() { cardImg = tempImg };

                //Bildauschnitt mit dem Kartenwert erzeugen 
                Crop cropValue = new Crop(new Rectangle(0, 0, 70, 60));
                Bitmap tempValueImg = cropValue.Apply(tempImg);
                //Wert-Ausschnitt nach S/W konvertieren
                tempValueImg = seq.Apply(tempValueImg);
                //Wert-Ausschnitt im Karten-Objekt ablegen
                card.identifierValueImg = tempValueImg;

                //Liste mit Wert-Samples erstellen
                List<Identifier> valueSamples = new List<Identifier>()
                {
                    new Identifier(){ name = "2", sample = KartenDetektor.Properties.Resources._2 },
                    new Identifier(){ name = "3", sample = KartenDetektor.Properties.Resources._3 },
                    new Identifier(){ name = "4", sample = KartenDetektor.Properties.Resources._4 },
                    new Identifier(){ name = "5", sample = KartenDetektor.Properties.Resources._5 },
                    new Identifier(){ name = "6", sample = KartenDetektor.Properties.Resources._6 },
                    new Identifier(){ name = "7", sample = KartenDetektor.Properties.Resources._7 },
                    new Identifier(){ name = "8", sample = KartenDetektor.Properties.Resources._8 },
                    new Identifier(){ name = "9", sample = KartenDetektor.Properties.Resources._9 },
                    new Identifier(){ name = "10", sample = KartenDetektor.Properties.Resources._10 },
                    new Identifier(){ name = "Bube", sample = KartenDetektor.Properties.Resources.j },
                    new Identifier(){ name = "Dame", sample = KartenDetektor.Properties.Resources.q },
                    new Identifier(){ name = "König", sample = KartenDetektor.Properties.Resources.k },
                    new Identifier(){ name = "Ass", sample = KartenDetektor.Properties.Resources._as }
                };

                //Alle Samples auf den Wert-Ausschnitt anwenden
                float maxSimilar = 0f;
                foreach (var valueSample in valueSamples)
                {
                    //Ähnlichkeit zwischen Sample und Wert-Ausschnitt ermitteln
                    ExhaustiveTemplateMatching tm = new ExhaustiveTemplateMatching(0);
                    TemplateMatch[] matchings = tm.ProcessImage(card.identifierValueImg, valueSample.sample);
                    
                    //Wenn das aktuell getestete Sample besser passt, als das bisher beste,
                    //setze den Wert als identifizierten Kartenwert
                    if (matchings.Length > 0 && matchings[0].Similarity > maxSimilar)
                    {
                        maxSimilar = matchings[0].Similarity;
                        card.value = valueSample.name;
                    }
                }

                //Bildauschnitt mit der Kartenfarbe erzeugen 
                Crop cropSuit = new Crop(new Rectangle(0, 60, 70, 65));
                Bitmap tempSuitImg = cropSuit.Apply(tempImg);
                //Kartenfarben-Ausschnitt nach S/W konvertieren
                tempSuitImg = seq.Apply(tempSuitImg);
                //Kartenfarben-Ausschnitt exakt zuschneiden und normalisieren
                tempSuitImg = normalizeForEqualityCheck(tempSuitImg);
                //Kartenfarben-Ausschnitt im Karten-Objekt ablegen
                card.identifierSuitImg = tempSuitImg;

                //Liste mit Kartenfarben-Samples erstellen
                List<Identifier> suitSamples = new List<Identifier>()
                    {
                        new Identifier(){ name = "Karo", sample = KartenDetektor.Properties.Resources.diamonds },
                        new Identifier(){ name = "Herz", sample = KartenDetektor.Properties.Resources.hearts },
                        new Identifier(){ name = "Pik", sample = KartenDetektor.Properties.Resources.spades },
                        new Identifier(){ name = "Kreuz", sample = KartenDetektor.Properties.Resources.clubs }
                    };

                //Alle Samples auf den Kartenfarben-Ausschnitt anwenden
                maxSimilar = 0f;
                foreach (var suitSample in suitSamples)
                {
                    //Ähnlichkeit zwischen Sample und Kartenfarben-Ausschnitt ermitteln
                    ExhaustiveTemplateMatching tm = new ExhaustiveTemplateMatching(0);
                    //Kartenfarben-Sample exakt zuschneiden, normalisieren und mit Ausschnitt abgleichen
                    TemplateMatch[] matchings = tm.ProcessImage(card.identifierSuitImg, normalizeForEqualityCheck(suitSample.sample));

                    //Wenn das aktuell getestete Sample besser passt, als das bisher beste,
                    //setze die Farbe als identifizierte Kartenfarbe
                    if (matchings.Length > 0 && matchings[0].Similarity > maxSimilar)
                    {
                        maxSimilar = matchings[0].Similarity;
                        card.suit = suitSample.name;
                    }
                }

                //Spielkarte der Spielkartenliste hinzufügen
                cards.Add(card);

                //Karte zur Visualisierung auf dem Screenshot markieren
                foreach (var c in edgePoints)
                {
                    gfx.DrawRectangle(Pens.Red, new Rectangle(c.X, c.Y, 1, 1));
                }
            }

            //Graphics-Objekt speichern, da nun alle Karten auf dem
            //Screenshot eingezeichnet sind.
            gfx.Save();

            //Kartenfarben-Ausschnitte in PictureBoxen ausgeben
            //und ermittelte Farben in Textboxen anzeigen
            pictureBoxSuit1.BackgroundImage = cards[0].identifierSuitImg;
            textBoxSuit1.Text = cards[0].suit;
            pictureBoxSuit2.BackgroundImage = cards[1].identifierSuitImg;
            textBoxSuit2.Text = cards[1].suit;
            pictureBoxSuit3.BackgroundImage = cards[2].identifierSuitImg;
            textBoxSuit3.Text = cards[2].suit;
            pictureBoxSuit4.BackgroundImage = cards[3].identifierSuitImg;
            textBoxSuit4.Text = cards[3].suit;
            pictureBoxSuit5.BackgroundImage = cards[4].identifierSuitImg;
            textBoxSuit5.Text = cards[4].suit;

            //Kartenwerte-Ausschnitte in PictureBoxen ausgeben
            //und ermittelte Werte in Textboxen anzeigen
            pictureBoxValue1.BackgroundImage = cards[0].identifierValueImg;
            textBoxValue1.Text = cards[0].value;
            pictureBoxValue2.BackgroundImage = cards[1].identifierValueImg;
            textBoxValue2.Text = cards[1].value;
            pictureBoxValue3.BackgroundImage = cards[2].identifierValueImg;
            textBoxValue3.Text = cards[2].value;
            pictureBoxValue4.BackgroundImage = cards[3].identifierValueImg;
            textBoxValue4.Text = cards[3].value;
            pictureBoxValue5.BackgroundImage = cards[4].identifierValueImg;
            textBoxValue5.Text = cards[4].value;

            //Karten-Grafiken in PictureBoxen ausgeben
            //und ermittelte Farben & Werte in Textboxen anzeigen
            pictureBoxCard1.BackgroundImage = cards[0].cardImg;
            textBoxCard1.Text = cards[0].suit + "-" + cards[0].value;
            pictureBoxCard2.BackgroundImage = cards[1].cardImg;
            textBoxCard2.Text = cards[1].suit + "-" + cards[1].value;
            pictureBoxCard3.BackgroundImage = cards[2].cardImg;
            textBoxCard3.Text = cards[2].suit + "-" + cards[2].value;
            pictureBoxCard4.BackgroundImage = cards[3].cardImg;
            textBoxCard4.Text = cards[3].suit + "-" + cards[3].value;
            pictureBoxCard5.BackgroundImage = cards[4].cardImg;
            textBoxCard5.Text = cards[4].suit + "-" + cards[4].value;

            //Screenshot mit eingezeichneten Karten und Arbeitskopie anzeigen
            pictureBoxScreenshot.BackgroundImage = inputBmp;
            pictureBoxScreenshotProcessed.BackgroundImage = processedBmp;
        }

        public static Bitmap normalizeForEqualityCheck(Bitmap source)
        {
            //Den ersten schwarzen Pixel von rechts erkennen (=linke Schnittkante)
            int left = -1, right = -1, up = -1, down = -1;
            int clip = 255;
            for (int x = 0; x < source.Width; x++)
            {
                for (int y = 0; y < source.Height; y++)
                {
                    var col = source.GetPixel(x, y);
                    if ((col.R + col.B + col.B) / 3f < clip)
                    {
                        left = x;
                        break;
                    }
                }
                if (left != -1)
                    break;
            }

            //Den ersten schwarzen Pixel von rechts erkennen (=rechte Schnittkante)
            for (int x = source.Width-1; x > 0; x--)
            {
                for (int y = 0; y < source.Height; y++)
                {
                    var col = source.GetPixel(x, y);
                    if ((col.R + col.B + col.B) / 3f < clip)
                    {
                        right = x;
                        break;
                    }
                }
                if (right != -1)
                    break;
            }

            //Den ersten schwarzen Pixel von oben erkennen (=obere Schnittkante)
            for (int y = 0; y < source.Height; y++)
            {
                for (int x = 0; x < source.Width; x++)
                {
                    var col = source.GetPixel(x, y);
                    if ((col.R + col.B + col.B) / 3f < clip)
                    {
                        up = y;
                        break;
                    }
                }
                if (up != -1)
                    break;
            }

            //Den ersten schwarzen Pixel von unten erkennen (=unten Schnittkante)
            for (int y = source.Height-1; y > 0; y--)
            {
                for (int x = 0; x < source.Width; x++)
                {
                    var col = source.GetPixel(x, y);
                    if ((col.R + col.B + col.B) / 3f < clip)
                    {
                        down = y;
                        break;
                    }
                }
                if (down != -1)
                    break;
            }
           
            //Grafik an den Schnittkanten zuschneiden
            Crop cropSuit = new Crop(new Rectangle(left, up, right-left, down-up));
            var output = cropSuit.Apply(source);
            //Grafik auf Norm-Maße strecken
            ResizeBilinear resizer = new ResizeBilinear(60, 60);
            return resizer.Apply(output);
        }
        
        //Spielkarten Klasse - enthält alle Informationen zu einer Karte
        private class Card
        {
            //Bild der Karte
            public Bitmap cardImg { get; set; }

            //Bildausschnitt mit der Kartenfarbe
            public Bitmap identifierSuitImg { get; set; }

            //Bildausschnitt mit dem Wert
            public Bitmap identifierValueImg { get; set; }

            //Analysierte Kartenfarbe
            public string suit { get; set; }

            //Analysierter Wert
            public string value { get; set; }
        }


        //Identitätsstruktur
        private struct Identifier
        {
            //Name des Samples (z.B. Karo, Ass, ...)
            public string name { get; set; }

            //Sample-Grafik
            public Bitmap sample { get; set; }
        }

        private void button1_Click(object sender, EventArgs e)
        {
            //Solange noch Screenshots vorhanden sind...
            if (files.Count() > 0)
            {
                //...lade Screenshot in Bitmap und starte Analyse
                Bitmap inputBmp = (Bitmap)Bitmap.FromFile(files[0]);
                analyzeCards(inputBmp);
                files.RemoveAt(0);
            }
            else
            {
                //Keine Screenshots mehr in der Liste
                MessageBox.Show("Alle Samples durch.");
            }
        }

    }
}
